/*--------------------------------------------------------------------
 *    The GMTPLUS-system:	@(#)cartergmtplus.c	1.2	9/12/95
 *
 *  Based on...
 *    The GMT-system:	@(#)mgd77togmt.c	2.6  7/23/93
 *
 *    Copyright (c) 1991 by P. Wessel and W. H. F. Smith
 *    See README file for copying and redistribution conditions.
 *--------------------------------------------------------------------*/
/*
 * cartermgtplus takes an existing gmtplus file and calculates
 * corrected meters from the two-way travel time.  A value in the
 * corrected meters column will be overwritten if there is a valid
 * travel time in the twt column.  If the twt column does not have
 * a valid time, the value in the corrected meters column will
 * remain unchanged.
 * 
 * This program will create new gmtplus files, and rename the
 * original files with a *.old extension.  Please allow sufficient
 * space for these additional files.  Additionally, pedigree lines
 * will be added to the header file to indicate this process has
 * been run, when, and by whom.
 *
 * The methodology used to calculate the corrected depth is that
 * used in mgd77togmt from the original gmt database program.  This
 * method was found to be correct.
 *
 * GMTPlus 03/10/95 jcm
 *
 */
#include <time.h>
#include "gmt.h"
#include "gmtplus_mgg.h"

#define NODATA -32000

#define N_BINS 64800
#define N_CARTER_ZONES 85
#define N_CARTER_OFFSETS 86
#define N_CARTER_CORRECTIONS 5812
#define MAX_VALID_TWT   16000   /* Two-way time (msec) to 12000 nominal m  */

int     carter_setup();
int     get_bin();
int     get_carter_zone();
int     get_depth_from_twt();
int     get_twt_from_depth();

short int       carter_zone[N_BINS];
short int       carter_offset[N_CARTER_OFFSETS];
short int       carter_correction[N_CARTER_CORRECTIONS];

struct GMT *gmt;
struct GMT_PLUS_REC *record;
struct COL_N col;
int T_INC = 60, fake_running_time = 0;


int	carter_not_initialized = 1;
int error_count = 0;
int bad_lon = 0;
int bad_lat = 0;
int bin_out_error = 0;
int zone_out_error = 0;
int bad_twt = 0;
int big_twt = 0;
int bad_depth = 0;
int big_depth = 0;

char path[80]; char *envdir;

main (argc, argv)
int argc;
char **argv; {
	int i, j, rec, n_records, n_alloc = GMT_CHUNK, leg_year;
	int t_flag = FALSE, error = FALSE, header = FALSE, reset_twtt_column = FALSE;
	char file[80], cruise[80] = "", agency[10] = "", legfname[80], line[512];
    char gmtfile[100], gmtheader[100];
	FILE *fp = NULL, *fpo, *hpo, *hp;
    int column;
    int verbose = 0;
    int nfiles = 0;
    char comment[80];
    long int    right_now;
    char    date[25], user[128], host[128];

    gmtpath_init();
    read_gmt_master_index();

    if (argc == 1 || error) {
      fprintf (stderr, "usage: cartergmtplus [-Pcarter_path]\n");
      fprintf(stderr,"	-P gives directory for the carter tables (default is $GMTHOME, then %s)\n", LIBDIR);
      
      exit (-1);
    }
    
    for (j=1; j<argc; j++) 
       if ((!strcmp(argv[j], "-v")) || (!strcmp(argv[j], "-V"))) verbose = TRUE;

    /* Give path to index files */
    for (j=1; j<argc; j++) 
      if ((!strcmp(argv[j], "-p")) || (!strcmp(argv[j], "-P"))) {
	strcpy (path, &argv[i][2]);
      }

    for (j=1; j<argc; j++) {  /* loop over gmt files */

       if (argv[j][0] == '-') continue;
        if (gmtpath_func (gmtfile, gmtheader, argv[j])) {
            fprintf(stderr,"cartergmtplus : Cannot find leg %s\n", argv[j]);
                continue;
        }
        if ((hp = fopen (gmtheader, "r")) == NULL) {
            fprintf(stderr,"cartergmtplus: Could not open header file %s\n", gmtheader);
            continue;
        }
        if ((fp = fopen (gmtfile, "r")) == NULL) {
            fprintf(stderr,"cartergmtplus: Could not open data file %s\n", gmtfile);
            fclose(hp);
            continue;
        }
        nfiles++;

	/* carter table path priority: 
	   1) -P argument
	   2) GMTPLUS_LIBDIR
	   3) default (LIBDIR)
	*/

	if (path[0] == 0) {
		envdir = getenv("GMTPLUS_LIBDIR");
		if (envdir == NULL) {
		  sprintf (path, "%s", LIBDIR);
		} else {
		  sprintf (path, "%s", envdir);
		}
	}
	/* printf("FIXME: looking for carter data in \"%s\"\n", path); */


        if (verbose) fprintf (stderr, "Now reading leg %s\n", argv[j]);

	read_gmt_header(hp, &col, &leg_year, &n_records, agency, cruise);
	read_first_gmt_datarec(fp, &col);
	if (col.twtt == 7 || col.tcor == 8) col.tcor = 8;	
	if (col.twtt == -1) {

		reset_twtt_column = TRUE;
		col.twtt = 7;
	}

        gmt = gmt_init (leg_year);  /* Initialize gmt_structure */


	record = (struct GMT_PLUS_REC *) GMT_memory ((char *)NULL, n_records, sizeof (struct GMT_PLUS_REC), "cartergmtplus");

    error_count = 0;
    bad_lon = 0;
    bad_lat = 0;
    bin_out_error = 0;
    zone_out_error = 0;
    bad_twt = 0;
    big_twt = 0;
    bad_depth = 0;
    big_depth = 0;

	for (rec=0; rec<n_records; rec++) {

       if (fread((char *)(&record[rec]), 40, 1, fp) != 1) {
          fprintf(stderr,"cartergmtplus: Fatal Error reading data record no. %d\n",rec);
          fprintf(stderr,"               File %s not changed!\n",gmtfile);
          free((char *)record);
          free((char *)gmt);
          fclose(fp);
          fclose(hp);
          continue;
       }

	   decode_record (&record[rec].ldata[col.lat],
           &record[rec].ldata[col.lon],
           &record[rec].sdata[col.twtt-6],
           &record[rec].sdata[col.tcor-6]);

	}
    fclose(hp);
    fclose(fp);
    sprintf(file,"%s.old\0",gmtheader);
    if (rename(gmtheader,file) != 0) {
       fprintf(stderr,"File rename failed: %s\n",file);
       exit(-1);
    }
    sprintf(file,"%s.old\0",gmtfile);
    if (rename(gmtfile,file) != 0) {
       fprintf(stderr,"File rename failed: %s\n",file);
       exit(-1);
    }
	
	if ((hpo = fopen (gmtheader, "w")) == NULL) {
		fprintf (stderr, "Could not create header file %s\n", gmtheader);
		exit (-1);
	}
	if ((fpo = fopen (gmtfile, "w")) == NULL) {
		fprintf (stderr, "Could not create data file %s\n", gmtfile);
        fclose(hpo);
		exit (-1);
	}

    /* assume no corrected data; fix if one was actually set */
    column = col.tcor;
    col.tcor = -1;
    for (rec=0; rec < n_records; rec++) {
       if (record[rec].sdata[column-6] != NODATA) {
          col.tcor = column;
          break;
       }
    }

	if (reset_twtt_column) col.twtt = -1;

	write_gmt_header(hpo, &col, leg_year, n_records, agency, cruise, gmtfile);
	write_first_gmt_datarec(fpo, &col);

	for (rec = 0; rec < n_records; rec++) {
		if (fwrite((char *)(&record[rec]), 40, 1, fpo) != 1) {
			fprintf(stderr,"cartergmtplus: Error writing data record no %d\n",rec);
			exit(-1);
		}
	}

	fclose(fpo);
	fclose(hpo);

    /* add pedigree lines */
    sprintf(file,"%s.old\0",gmtheader);
    hp = fopen(file,"r");
    hpo = fopen(gmtheader,"a");
    transfer_header_comments(hp, hpo);

    add_pedigree(hpo,"#FT: gmtplus by cartergmtplus");

    strncpy(comment,"\0",80);
    sprintf(comment,"#FN: %s",gmtheader);
    add_pedigree(hpo,comment);

    strncpy(date,"\0",24);
    right_now = time((long *)0);
    strcpy(date,ctime(&right_now));
    date[24] = 0; /* punt the nl */
    strcpy(user,getenv("USER"));
    gethostname(host,128);

    strncpy(comment,"\0",80);
    sprintf(comment,"#FD:%s",date);
    add_pedigree(hpo, comment);

    strncpy(comment,"\0",80);
    sprintf(comment,"#HU:%s %s",host,user);
    add_pedigree(hpo, comment);

    add_pedigree(hpo,"#*end");
    fclose(hp);
    fclose(hpo);
 
    free((char *)gmt);	
	free ((char *)record);

       if (error_count != 0) {
          fprintf(stderr,"%s: %d errors\n",gmtfile,error_count);
          if (bad_lon != 0)
             fprintf(stderr,"\t%d longitude domain errors\n",bad_lon);
          if (bad_lat != 0)
             fprintf(stderr,"\t%d latitude domain errors\n",bad_lat);
          if (bin_out_error != 0)
             fprintf(stderr,"\t%d Bin out-of-range\n",bin_out_error);
          if (zone_out_error != 0)
             fprintf(stderr,"\t%d Zone out-of-range\n",zone_out_error);
          if (bad_twt != 0)
             fprintf(stderr,"\t%d Negative travel time\n",bad_twt);
          if (big_twt != 0)
             fprintf(stderr,"\t%d Travel time too large\n",big_twt);
          if (bad_depth != 0)
             fprintf(stderr,"\t%d Negative depth\n",bad_depth);
          if (big_depth != 0)
             fprintf(stderr,"\t%d Depth too large\n",big_depth);
       }

    }

    if (nfiles == 0) {
        fprintf(stderr, "usage: cartergmtplus leg(s)\n");
        exit(-1);
    }

	
	exit(0);
}

decode_record (lat,lon,itwt, itcor)
int *lat, *lon;
short int *itwt, *itcor;
{
	int year, month, day, hour, min, sec, l_mag, l_top, test, bin, zone, l_twt;
	short int	twt;
	double tz, fmin;
	char s_tz[5], s_yr[3], s_mo[3], s_dy[3], s_hr[3], s_mi[6], s_lat[8], s_lon[9];
	char s_faa[10], s_mag[10], s_top[7], s_top_twt[7];
	short int *top;
	short int top_cheat;

	top = &top_cheat;

    /* two-way travel time */

	twt = *itwt;
	if (twt != NODATA) {	

		if ((get_bin_6(*lat, *lon, &bin)) || (get_carter_zone(bin, &zone)) || (get_depth_from_twt(zone, twt, itcor)) ) {
			fprintf(stderr,"cartergmtplus:  ERROR in Carter correction system at lat/lon/twtt:\t%d\t%d\t%d\n", *lat/10, *lon/10, *itwt);
			*itcor = NODATA;	
		} 
		else *itcor = -(*itcor);
	}
}

/* The following carter depth calculation code is
   substantially the same as from the original mgd77togmt.c code
   by Walter Smith et al.
*/

int	get_bin_6(lat,lon, bin)
int	lat;
int	lon;
int	*bin;
{
	/* Given signed long ints in the 1.0e06 times decimal degree range, 
		-90000000 <= lat < 90000000, 0 <= lon < 360000000, set bin
		number.
		Returns 0 if OK, -1 if error.  */

	int	latdeg, londeg;

	if (lat < -90000000 || lat > 90000000) {
		/*fprintf(stderr, "getbin_6:  Latitude domain error.\n");*/
        error_count++;
        bad_lat++;
		return(-1);
	}
	if (lon < 0 || lon > 360000000) {
		/*fprintf(stderr, "getbin_6:  Longitude domain error.\n");*/
        error_count++;
        bad_lon++;
		return(-1);
	}
	latdeg = (lat + 90000000)/1000000;
	if (latdeg == 180) latdeg = 179;	/* Map north pole to previous row  */

	londeg = lon/1000000;
	*bin = 360 * latdeg + londeg;

	return(0);
}

int	carter_setup()
{
	/* This routine must be called once before using carter table stuff.
	It reads the carter_corr.b, carter_offset.b, carter_zone.b files and
	loads the appropriate arrays.  It sets carter_not_initialized = FALSE
	upon successful completion and returns 0.  If failure occurrs, it
	returns -1.  */

	FILE	*fp = NULL;
	char	buffer [512];
	int	i;

	/* Read the correction table:  */

	sprintf(buffer,"%s/carter_corr.b", path);
	if ( (fp = fopen(buffer, "r")) == NULL) {
                fprintf(stderr,"carter_setup:  Cannot open r %s\n", buffer);
		carter_not_initialized = TRUE;
                return(-1);
        }
        for (i = 0; i < N_CARTER_CORRECTIONS; i++) {
                if ((fread((char *)&carter_correction[i], sizeof(short int), 1, fp)) != 1) {
                        fprintf(stderr, "carter_setup:  Cannot read record %ld from %s\n", i, buffer);
		carter_not_initialized = TRUE;
                return(-1);
                }
        }
        fclose(fp);


	/* Read the offset table:  */

	sprintf(buffer,"%s/carter_offset.b", path);
	if ( (fp = fopen(buffer, "r")) == NULL) {
                fprintf(stderr,"carter_setup:  Cannot open r %s\n", buffer);
		carter_not_initialized = TRUE;
                return(-1);
        }
        for (i = 0; i < N_CARTER_OFFSETS; i++) {
                if ((fread((char *)&carter_offset[i], sizeof(short int), 1, fp)) != 1) {
                        fprintf(stderr, "carter_setup:  Cannot read record %ld from %s\n", i, buffer);
		carter_not_initialized = TRUE;
                return(-1);
                }
        }
        fclose(fp);


	/* Read the zone table:  */

	sprintf(buffer,"%s/carter_zone.b", path);
	if ( (fp = fopen(buffer, "r")) == NULL) {
                fprintf(stderr,"carter_setup:  Cannot open r %s\n", buffer);
		carter_not_initialized = TRUE;
                return(-1);
        }
        for (i = 0; i < N_BINS; i++) {
                if ((fread((char *)&carter_zone[i], sizeof(short int), 1, fp)) != 1) {
                        fprintf(stderr, "carter_setup:  Cannot read record %ld from %s\n", i, buffer);
		carter_not_initialized = TRUE;
                return(-1);
                }
        }
        fclose(fp);

	/* Get here when all is well.  */

	carter_not_initialized = FALSE;
	return(0);
}


int	get_carter_zone(bin, zone)
int	bin;
int	*zone;
{
	/* Sets value pointed to by zone to the Carter zone corresponding to
		the bin "bin".  Returns 0 if successful, -1 if bin out of
		range.  */

	if (carter_not_initialized && carter_setup() ) {
		fprintf(stderr,"get_carter_zone:  Initialization failure.\n");
		return(-1);
	}

	if (bin < 0 || bin >= N_BINS) {
		/*fprintf(stderr,"get_carter_zone:  bin out of range.\n");*/
        error_count++;
        bin_out_error++;
		return(-1);
	}
	*zone = carter_zone[bin];
	return(0);
}


int	get_depth_from_twt(zone, twt_in_msec, depth_in_corr_m)
int	zone;
short int	twt_in_msec;
short int *	depth_in_corr_m;
{
	/* Given two-way travel time of echosounder in milliseconds, and
		Carter Zone number, finds depth in Carter corrected meters.
		Returns (0) if OK, -1 if error condition.  */

	int	i, nominal_z1500, low_hundred, part_in_100;

	if (carter_not_initialized && carter_setup() ) {
		fprintf(stderr,"get_depth_from_twt:  Initialization failure.\n");
		return(-1);
	}
	if (zone < 1 || zone > N_CARTER_ZONES) {
		/*fprintf(stderr,"get_depth_from_twt:  Zone out of range.\n");*/
        error_count++;
        zone_out_error++;
		return(-1);
	}
	if (twt_in_msec < 0) {
		/*fprintf(stderr,"get_depth_from_twt:  Negative twt.\n");*/
        error_count++;
        bad_twt++;
		return(-1);
	}

	nominal_z1500 = irint(0.75 * twt_in_msec);

	if (nominal_z1500 <= 100) {
		/* There is no correction in water this shallow.  */
		*depth_in_corr_m = nominal_z1500;
		return(0);
	}

	low_hundred = nominal_z1500 / 100;
	i = carter_offset[zone - 1] + low_hundred - 1;	/* -1 'cause .f indices */
	
	if (i >= (carter_offset[zone] - 1) ) {
		/*fprintf(stderr, "get_depth_from_twt:  twt too big.\n");*/
        error_count++;
        big_twt++;
		return(-1);
	}

	part_in_100 = nominal_z1500%100;

	if (part_in_100) {	/* We have to interpolate the table  */

		if ( i == (carter_offset[zone] - 2) ) {
			/*fprintf(stderr, "get_depth_from_twt:  twt too big.\n");*/
            error_count++;
            big_twt++;
			return(-1);
		}

		*depth_in_corr_m = (short) irint (carter_correction[i] +
			0.01 * part_in_100 * (carter_correction[i+1]
				- carter_correction[i]) );
		return(0);
	}
	else {
		*depth_in_corr_m = carter_correction[i];
		return(0);
	}
}


int	get_twt_from_depth(zone, depth_in_corr_m, twt_in_msec)
int	zone;
short int	depth_in_corr_m;
short int *	twt_in_msec;
{
	/*  Given Carter zone and depth in Carter corrected meters,
	finds the two-way travel time of the echosounder in milliseconds.
	Returns -1 upon error, 0 upon success.  */

	int	min, max, guess;
	double	fraction;

	if (carter_not_initialized && carter_setup() ) {
		fprintf(stderr,"get_twt_from_depth:  Initialization failure.\n");
		return(-1);
	}
	if (zone < 1 || zone > N_CARTER_ZONES) {
		/*fprintf(stderr,"get_twt_from_depth:  Zone out of range.\n");*/
        error_count++;
        zone_out_error++;
		return(-1);
	}
	if (depth_in_corr_m < 0) {
		/*fprintf(stderr,"get_twt_from_depth:  Negative depth.\n");*/
        error_count++;
        bad_depth++;
		return(-1);
	}

	if (depth_in_corr_m <= 100) {	/* No correction applies.  */
		*twt_in_msec = irint(1.33333 * depth_in_corr_m);
		return(0);
	}

	max = carter_offset[zone] - 2;
	min = carter_offset[zone - 1] - 1;

	if (depth_in_corr_m > carter_correction[max]) {
		/*fprintf(stderr,"get_twt_from_depth:  Depth too big.\n");*/
        error_count++;
        big_depth++;
		return(-1);
	}

	if (depth_in_corr_m == carter_correction[max]) {
		/* Hit last entry in table exactly  */
		*twt_in_msec = irint(133.333 * (max - min) );
		return(0);
	}


	guess = (depth_in_corr_m / 100) + min;
	if (guess > max) guess == max;
	while (guess < max && carter_correction[guess] < depth_in_corr_m) guess++;
	while (guess > min && carter_correction[guess] > depth_in_corr_m) guess--;

	if (depth_in_corr_m == carter_correction[guess]) {
		/* Hit a table value exactly  */
		*twt_in_msec = irint(133.333 * (guess - min) );
		return(0);
	}
	fraction = ( (double)(depth_in_corr_m - carter_correction[guess])
			/ (double)(carter_correction[guess + 1] - carter_correction[guess]) );
	*twt_in_msec = irint(133.333 * (guess - min + fraction) );
	return(0);
}

