/*--------------------------------------------------------------------
 *    The GMTPLUS-system:	@(#)gmtplus2bin.c	1.2	9/12/95
 *
 *  Based on...
 *    The GMT-system:	@(#)gmt2bin.c	2.4  7/22/93
 *
 *    Copyright (c) 1991 by P. Wessel and W. H. F. Smith
 *    See README file for copying and redistribution conditions.
 *--------------------------------------------------------------------*/
/*
 * gmt2bin
 * program to read a leg list, open legid.gmt files, and determine what 1 x 1
 * degree bins of the earth are occupied by each leg and what data types exist
 * in each bin.  For each leg input, program writes a legid.bix file.
 * Note:  The program asks at the start if the legs in the leg list are inside
 * or outside legs; all legs in the list MUST be one or the other.  We flag the
 * bin_index file with this for each cruise in this session.
 *
 *	Input file :  a legid.gmt file, direct, recl = 18, the standard file
 *			of W. H. F. Smith and P. Wessel.  See makegmt.f
 *
 *	Output file:  a legid.bix file, as follows:
 *		first record: year1, total_gmt_flags (sum of 1(G), 2(M), 4(T)
 *		remaining recs: bindex gmt_flags 
 *
 *		year1 is the year of the first point in the file
 *		total_gmt_flags for existance of data over whole cruise.
 *
 *	New version 2.0	13-Jun-1991	P. Wessel
 *
 * GMTPlus version 03/10/95 jcm
 * Modifications made for gmtplus database system.
 *
 * gmtplus2bin uses the same algorithm as in gmt2bin to determine inclusion
 * in a bin, bin number, etc.
 *
 * Input file: standard gmtplus {legid.gmtd, legid.gmth} pair
 *
 * Output file: legid.bix, as follows:
 *     first record: year of first data record, followed by list of
 *                   data types total in the file.
 *     remaining recs: bindex, followed by list of data types in the
 *                   specific bin.
 *   Note: lat, lon and time are assumed to exist in all files and are thus
 *         not listed in the database index.
 *
 */

#include <gmt.h>
#include "gmtplus_mgg.h"

#define MEM_SIZE 100000
#define MEM_INC 50000
#define MBINS 64800
#define NODATA -32000

struct GMT_PLUS_REC *record;
struct COL_N col;

int
main (int argc, char *argv[])
{
	int bindex, million = 1000000, mil90, error = FALSE;
	int year, n_records, lastb, llat1, llat2, llon1, llon2, newbin, flag;
	int i, j;
	double xedge, yedge, xx1, xx2, yy1, yy2, delx, dely, slope, xedgey, yedgex, rxedge, ryedge;
	char legid[16], agency[10] = "", gmtfile[80], gmtheader[80], binfile[80];
	char cruise[80] = "";
	int inboxn[MBINS];
	char line[512], path[80];
	FILE *fpl, *fpo, *fp, *hp;
	int boxflags[MBINS][14];
	bin_flag_ptr inboxgmt[MBINS], totalgmt;
	bin_flag_ptr head, tail;
	int nfiles = 0;

	path[0] = 0;
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'V':
					gmtdefs.verbose = TRUE;
					break;
				case 'P':
					strcpy (path, &argv[i][2]);
					break;
				default:
					error = TRUE;
					break;
			}
		}
		else if ((fpl = fopen (argv[i], "r")) == NULL) {
			fprintf (stderr, "gmtplus2bin: Couldn't open file %s\n", argv[i]);
			exit (-1);
		}
	}
	
	if (argc == 1 || error) {
		fprintf (stderr, "usage: gmtplus2bin leglistfile [-Ppath -V]\n");
		fprintf (stderr, "	-P full path to directory with gmtfiles [default is current directory]\n");
		fprintf (stderr, "	-V means verbose\n");
		exit (-1);
	}

	read_gmt_master_index();
	
	record = (struct GMT_PLUS_REC *) GMT_memory ((char *)NULL, 1, sizeof (struct GMT_PLUS_REC), "gmtplus2bin");
	
	mil90 = million * 90;
	
	while (fgets (line, 512, fpl)) {
		sscanf (line, "%s", legid);
		if (path[0]) {
			sprintf (gmtfile, "%s/%s.gmtd\0", path, legid);
			sprintf (gmtheader, "%s/%s.gmth\0", path, legid);
		} else {
			sprintf (gmtfile, "%s.gmtd\0", legid);
			sprintf (gmtheader, "%s.gmth\0", legid);
		}
		if ((fp = fopen (gmtfile, "r")) == NULL) {
			fprintf (stderr, "gmtplus2bin: Cannot open file %s\n", gmtfile);
			continue;
		}
		if ((hp = fopen (gmtheader, "r")) == NULL) {
			fprintf (stderr, "gmtplus2bin: Cannot open file %s\n", gmtheader);
			continue;
		}
		if (gmtdefs.verbose) fprintf (stderr, "Processing leg %s\n", legid);
		nfiles++;
		read_gmt_header(hp, &col, &year, &n_records, agency, cruise);
		read_first_gmt_datarec(fp, &col);
		
		record = (struct GMT_PLUS_REC *) GMT_memory ((char *)record, n_records, sizeof (struct GMT_PLUS_REC), "gmtplus2bin");
		
		i = 0;
		/*while (fread ((char *)&record[i], 40, 1, fp)) {*/
		/* 08/31/95 jcm do not include bad data or flags in database */
		while (read_data_record(&record[i], &col, fp) == 0) {
			
			while (record[i].ldata[col.lon] < 0) record[i].ldata[col.lon] += 360000000;
			if (record[i].ldata[col.lon] > 360000000) continue;
			if (record[i].ldata[col.lat] < -90000000 
				|| record[i].ldata[col.lat] > 90000000) continue;
			record[i].ldata[col.lat] += mil90;
			i++;
		}
		fclose (fp);
		fclose(hp);
		n_records = i;
		
		totalgmt = NULL;
		for (i = 0; i < MBINS; i++) {
			inboxn[i] = 0;
			inboxgmt[i] = NULL;
			for (j=0; j<14; j++)
				boxflags[i][j] = 0;
		}
		/* 
		 * find occupied boxes, and also 'implied' occupancy for later xover checks. 
		 * an implied box is one whose corner was cut by a ship going from one bin to
		 * another; we find these only when the deltat between points is < 15 minutes,
		 * so crossover interpolation applies.  This should arise only rarely.  The 15
		 * minute stipulation also means that in such cases there should be only one
		 * implied box; the following code tries to find only one.
		 */
		
		bindex = (record[0].ldata[col.lat] / million) * 360 + record[0].ldata[col.lon] / million;
		if (bindex >= MBINS) {
			fprintf (stderr, "gmtplus2bin: Index too big! (%d)\n", bindex);
			exit (-1);
		}
		lastb = bindex;
		inboxn[bindex] = 1;
		
		/* keep track of total types of data */
		flag = 0;
		head = totalgmt;
		for (i=0;i<6;i++) {
			if ((first_record.ldata[i] != -1) 
				&& (col.tsec != i)
				&& (col.lat != i)
				&& (col.lon != i)
				&& (col.bad_flag != i)) {
				if (head == NULL) {
					head = totalgmt = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
				} else {
					head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
					head = head->next;
				}
				head->next = NULL;
				head->data = first_record.ldata[i];
			}
		}
		for (i=0;i<8;i++) {
			if ((first_record.sdata[i] != -1) 
				&& (col.tsec != i+6)
				&& (col.lat != i+6)
				&& (col.lon != i+6)
				&& (col.bad_flag != i+6)) {
				if (head == NULL) {
					head = totalgmt = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
				} else {
					head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
					head = head->next;
				}
				head->next = NULL;
				head->data = first_record.sdata[i];
			}
		}
		/* record data particular to this bindex */
		head = inboxgmt[bindex];
		for (i=0; i<6; i++) {
			if (record[0].ldata[i] != NODATA
				&& (col.tsec != i)
				&& (col.lat != i)
				&& (col.lon != i)
				&& (col.bad_flag != i)
				&& (first_record.ldata[i] != -1)) {
				if (head == NULL) {
					head = inboxgmt[bindex] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
				} else {
					head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
					head = head->next;
				}
				head->next = NULL;
				head->data = first_record.ldata[i];
				boxflags[bindex][i] = 1;
			}
		}
		for (i=0; i<8; i++) {
			if (record[0].sdata[i] != NODATA
				&& (col.tsec != (i+6))
				&& (col.lat != (i+6))
				&& (col.lon != (i+6))
				&& (col.bad_flag != (i+6))
				&& (first_record.sdata[i] != -1)) {
				if (head == NULL) {
					head = inboxgmt[bindex] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
				} else {
					head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
					head = head->next;
				}
				head->next = NULL;
				head->data = first_record.sdata[i];
				boxflags[bindex][i+6] = 1;
			}
		}
		
		for (i = 1; i < n_records; i++) {
			bindex = (record[i].ldata[col.lat] / million) * 360 
				+ record[i].ldata[col.lon] / million;
			if (bindex >= MBINS) {
				fprintf (stderr, "gmtplus2bin: Index too big! (%d)\n", bindex);
				exit (-1);
			}

			if (inboxn[bindex] == 0) { /* first record */
				inboxn[bindex] = 1;
				head = inboxgmt[bindex];
				for (j=0; j<6; j++) {
					if (record[i].ldata[j] != NODATA
						&& (col.tsec != j)
						&& (col.lat != j)
						&& (col.lon != j) 
						&& (col.bad_flag != j)
						&& (first_record.ldata[j] != -1)) {
						if (head == NULL) {
							head = inboxgmt[bindex] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
						} else {
							head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
							head = head->next;
						}
						head->next = NULL;
						head->data = first_record.ldata[j];
						boxflags[bindex][j] = 1;
					}
				}
				for (j=0; j<8; j++) {
					if (record[i].sdata[j] != NODATA
						&& (col.tsec != (j+6))
						&& (col.lat != (j+6))
						&& (col.lon != (j+6))
						&& (col.bad_flag != (j+6))
						&& (first_record.sdata[j] != -1)) {
						if (head == NULL) {
							head = inboxgmt[bindex] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
						} else {
							head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
							head = head->next;
						}
						head->next = NULL;
						head->data = first_record.sdata[j];
						boxflags[bindex][j+6] = 1;
					}
				}
			} else { /* check for inclusion */
				head = inboxgmt[bindex];
				if (head != NULL)
					while (head->next != NULL)
						head = head->next;
				for (j=0; j<6; j++) {
					if (record[i].ldata[j] != NODATA
						&& (boxflags[bindex][j] == 0) /* not already in bin */
						&& (col.tsec != j)
						&& (col.lat != j)
						&& (col.lon != j)
						&& (col.bad_flag != j)
						&& (first_record.ldata[j] != -1)) {
						if (head == NULL) {
							head = inboxgmt[bindex] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
						} else {
							head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
							head = head->next;
						}
						head->next = NULL;
						head->data = first_record.ldata[j];
						boxflags[bindex][j] = 1;
					}
				}
				for (j=0; j<8; j++) {
					if (record[i].sdata[j] != NODATA
						&& (boxflags[bindex][j+6] == 0) /* not already in bin */
						&& (col.tsec != (j+6))
						&& (col.lat != (j+6))
						&& (col.lon != (j+6))
						&& (col.bad_flag != (j+6))
						&& (first_record.sdata[j] != -1)) {
						if (head == NULL) {
							head = inboxgmt[bindex] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
						} else {
							head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
							head = head->next;
						}
						head->next = NULL;
						head->data = first_record.sdata[j];
						boxflags[bindex][j+6] = 1;
					}
				}
			}
					
			if (lastb != bindex && 
				(record[i].ldata[col.tsec] - record[i-1].ldata[col.tsec]) <= 900) {
				yy1 = record[i-1].ldata[col.lat] * 0.000001;
				yy2 = record[i].ldata[col.lat] * 0.000001;
 				xx1 = record[i-1].ldata[col.lon] * 0.000001;
				xx2 = record[i].ldata[col.lon] * 0.000001;
				llat1 = yy1;
				llat2 = yy2;
				llon1 = xx1;
				llon2 = xx2;
				if (llat1 != llat2 && llon1 != llon2) {
					dely = yy2 - yy1;
					yedge = (dely > 0.0) ? yy2 : yy1;
					delx = xx2 - xx1;
					if (fabs (delx) > 180.0) {
						if (delx > 180.0) {
							delx -= 360.0;
							xx1 += 360.0;
							xedge = xx1;
						}
						else {
							delx += 360.0;
							xx2 += 360.0;
							xedge = xx2;
						}
					}
					else
						xedge = (delx > 0.0) ? xx2 : xx1;
					slope = dely / delx;
					xedgey = slope * (xedge - xx1) + yy1;
					yedgex = ( (yedge - yy1) + slope * xx1 ) / slope;
					rxedge = hypot (xedge - xx1, xedgey - yy1);
					ryedge = hypot (yedgex - xx1, yedge - yy1);
					if (rxedge > ryedge)
						newbin = (record[i].ldata[col.lat] / million) * 360 
								+ record[i-1].ldata[col.lon] / million;
					else
						newbin = (record[i-1].ldata[col.lat] / million) * 360 
								+ record[i].ldata[col.lon] / million;
					if (newbin >= MBINS) {
						fprintf (stderr, "gmtplus2bin: Index too big! (%d)\n", bindex);
						exit (-1);
					}
					if (inboxn[newbin] == 0) { /* first record */
						inboxn[newbin] = 1;
						head = inboxgmt[newbin];
						for (j=0; j<6; j++) {
							if (record[i].ldata[j] != NODATA
								&& (col.tsec != j)
								&& (col.lat != j)
								&& (col.lon != j) 
								&& (col.bad_flag != j)
								&& (first_record.ldata[j] != -1)) {
								if (head == NULL) {
									head = inboxgmt[newbin] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
								} else {
									head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
									head = head->next;
								}
								head->next = NULL;
								head->data = first_record.ldata[j];
								boxflags[newbin][j] = 1;
							}
						}
						for (j=0; j<8; j++) {
							if (record[i].sdata[j] != NODATA
								&& (col.tsec != (j+6))
								&& (col.lat != (j+6))
								&& (col.lon != (j+6))
								&& (col.bad_flag != (j+6))
								&& (first_record.sdata[j] != -1)) {
								if (head == NULL) {
									head = inboxgmt[newbin] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
								} else {
									head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
									head = head->next;
								}
								head->next = NULL;
								head->data = first_record.sdata[j];
								boxflags[newbin][j+6] = 1;
							}
						}
					} else { /* check for inclusion */
						head = inboxgmt[newbin];
						if (head != NULL)
							while (head->next != NULL)
								head = head->next;
						for (j=0; j<6; j++) {
							if (record[i].ldata[j] != NODATA
								&& (boxflags[newbin][j] == 0) /* not already in bin */
								&& (col.tsec != j)
								&& (col.lat != j)
								&& (col.lon != j)
								&& (col.bad_flag != j)
								&& (first_record.ldata[j] != -1)) {
								if (head == NULL) {
									head = inboxgmt[newbin] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
								} else {
									head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
									head = head->next;
								}
								head->next = NULL;
								head->data = first_record.ldata[j];
								boxflags[newbin][j] = 1;
							}
						}
						for (j=0; j<8; j++) {
							if (record[i].sdata[j] != NODATA
								&& (boxflags[newbin][j+6] == 0) /* not already in bin */
								&& (col.tsec != (j+6))
								&& (col.lat != (j+6))
								&& (col.lon != (j+6))
								&& (col.bad_flag != (j+6))
								&& (first_record.sdata[j] != -1)) {
								if (head == NULL) {
									head = inboxgmt[newbin] = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
								} else {
									head->next = (bin_flag_ptr)calloc(1, sizeof(bin_flag_type));
									head = head->next;
								}
								head->next = NULL;
								head->data = first_record.sdata[j];
								boxflags[newbin][j+6] = 1;
							}
						}
					}
				}
			}
			lastb = bindex;
		}
	
		/*
		 * new feature:
		 */
	 
		if (totalgmt == NULL)
			fprintf (stderr, "%s has empty data fields\n", gmtfile);
		else {
			sprintf (binfile, "%s.bix\0", legid);
			fpo = fopen (binfile, "w");
			fprintf (fpo, "%d", year);
			head = totalgmt;
			while (head != NULL) {
				fprintf(fpo, " %d", head->data);
				tail = head;
				head = head->next;
				free(tail);
			}
			fprintf(fpo, "\n");
			for (i = 0; i < MBINS; i++) {
				if (inboxn[i] && (inboxgmt[i] != NULL)) {
					fprintf (fpo, "%d ", i);
					head = inboxgmt[i];
					while (head != NULL) {
						fprintf(fpo, "%d ", head->data);
						tail = head;
						head = head->next;
						free(tail);
					}
					fprintf(fpo, "\n");
					fflush(fpo);
				}
			}
			fclose (fpo);
		}
	}
	fclose (fpl);
	
	free ((char *)record);

	if (nfiles == 0) {
		fprintf (stderr, "usage: gmtplus2bin leglistfile [-Ppath -V]\n");
		fprintf (stderr, "	-P full path to directory with gmtfiles [default is current directory]\n");
		fprintf (stderr, "	-V means verbose\n");
		exit (-1);
	}
	
	exit (0);
}
