/*--------------------------------------------------------------------
 *    The GMTPLUS-system:	@(#)gmtpluslegs.c	1.1	3/24/95
 *
 *  Based on...
 *	The GMT-system:	@(#)gmtlegs.c	2.4  7/20/93
 *
 *	Copyright (c) 1991 by P. Wessel and W. H. F. Smith
 *	See README file for copying and redistribution conditions.
 *--------------------------------------------------------------------*/
/*
 * gmtlegs - A program that locates all legs within a given
 * area, or only those legs that have certain types of data.
 * The program differs from LOC on VAX in that it checks each
 * 1 by 1 degree block to see if the leg actually crossed it.
 *
 * Author:	Paul Wessel
 * Date:	25-MAY-1987
 * Revised:	13-FEB-1991
 * Version 2.0: 21-Jun-1991
 *
 * GMTPlus version 03/10/95 jcm
 * Modifications made for gmtplus database system.
 * Note: routines dealing with INFO structure moved to gmtplus_mgg.c
 *
 */

#include <stdlib.h> 
#include <gmt.h>
#include "gmtplus_mgg.h"

#define MSIZE 64000

int inside (double w, double e, double s, double n, int binnr)
{
	float lat, lon;
	lon = (float) (binnr%360);
	while (lon > e) lon -= 360.0;
	if (lon < w || lon >= e) return (FALSE);
	lat = (float)(binnr/360) - 90.0;
	if (lat < s || lat >= n) return (FALSE);
	return (TRUE);
}


struct INFO *make_info();

int
main (int argc, char *argv[])
{
	int i, j, ninfo, n_found = 0, bin, leg, no, flagin[14], nlegs;
	int num, success;
	bin_flag_ptr gmt, gmtflag;
	char lname[16], line[512], dfile[512], bfile[512],path[120];
	char *envdir;
	double west, east, south, north;
	FILE *fleg, *fbin;
	int error = FALSE, full = FALSE, verbose = FALSE;
	struct INFO *leginfo[MSIZE], *ptr[MSIZE];
	int number;
	
	west = east = south = north = 0.0;
  	gmt = gmtflag = NULL;
  	
	for (i = 1; i < argc; i++) {
  		if (argv[i][0] == '-') {
			if (gmtflag == NULL) {
				gmt = gmtflag = (bin_flag_ptr)calloc(1,sizeof(bin_flag_type));
			} else {
				if (gmt->data != -9999) {
					gmt->next = (bin_flag_ptr)calloc(1,sizeof(bin_flag_type));
					gmt = gmt->next;
				}
			}
			gmt->data = -9999;
			gmt->next = NULL;
			switch (argv[i][1]) {
			case 'R':
			case 'V':
			case '\0':
				error += GMT_get_common_args (argv[i], &west, &east, &south, &north);
				break;

				case 'P':      /* give path to index files */
					strcpy(path, &argv[i][2]);
					break; 
				case 'G':   
				case 'g':	   /* Gravity is requested */
					switch (argv[i][2]) {
						case 'R':
						case 'r':
							gmt->data = GRAW;
							break;
						case 'F':
						case 'f':
							gmt->data = GFAA;
							break;
						default:
							error = TRUE;
							break;
					}
					break;
				case 'M':   
				case 'm':	   /* Magnetics is requested */
					switch (argv[i][2]) {
						case 'T':
						case 't':
							gmt->data = MTOT;
							break;
						case 'A':
						case 'a':
							gmt->data = MANM;
							break;
						case 'F':
						case 'f':
							gmt->data = MFILT;
							break;
						default:
							error = TRUE;
							break;
					}
                    break;
				case 'T':  
				case 't':	   /* Topography is requested */
					switch (argv[i][2]) {
						case 'T':
						case 't':
							gmt->data = TWTT;
							break;
						case 'C':
						case 'c':
							gmt->data = TCORR;
							break;
						default:
							error = TRUE;
							break;
					}
					break;
				case 'U':
				case 'u':
					switch (argv[i][2]) {
						case 'L':
						case 'l':
							switch (argv[i][3]) {
								case '0':
									gmt->data = UL0;
									break;
								case '1':
									gmt->data = UL1;
									break;
								case '2':
									gmt->data = UL2;
									break;
								case '3':
									gmt->data = UL3;
									break;
								case '4':
									gmt->data = UL4;
									break;
								case '5':
									gmt->data = UL5;
									break;
								default:
									error = TRUE;
									break;
							}
							break;
						case 'S':
						case 's':
							switch (argv[i][3]) {
								case '0':
									gmt->data = US0;
									break;
								case '1':
									gmt->data = US1;
									break;
								case '2':
									gmt->data = US2;
									break;
								case '3':
									gmt->data = US3;
									break;
								case '4':
									gmt->data = US4;
									break;
								case '5':
									gmt->data = US5;
									break;
								case '6':
									gmt->data = US6;
									break;
								case '7':
									gmt->data = US7;
									break;
								default:
									error = TRUE;
									break;
							}
							break;
						default:
							error = TRUE;
							break;
					}
					break;
				
  				case 'L':
  					full = TRUE;
  					break;
  				default:
  					error = TRUE;
  					break;
  			}
  		}
  		else
  			error = TRUE;
  	}
	gmt = gmtflag;
	if (gmt != NULL) {
		if (gmt->data == -9999) {
			free(gmtflag);
			gmtflag = NULL;
		} else {
			while (gmt->next != NULL) {
				if (gmt->next->data == -9999) {
					free(gmt->next);
					gmt->next = NULL;
				} else {
					gmt = gmt->next;
				}
			}
		}
	}

  	while (west > east) west -= 360.0;
	if (east > 360.0) east -= 360.0;
  	if (south >= north) error = TRUE;
  	if (west >= east) error = TRUE;
  	if (argc == 1 || error) {
  		fprintf(stderr,"usage: gmtpluslegs -R<west>/<east>/<south>/<north> [-P<path>] [-L] [-GR] [-GF] [-MT] [-MA] [-MF] [-TT] [-TC] [-V]\n");
  		fprintf(stderr,"	-R <west>, <east>, <south>, <north> is region in degrees.\n");
  		fprintf(stderr,"	-P<path> gives directory for index files [Default is %s].\n",LIBDIR);
  		fprintf(stderr,"	Choose any combination of:\n");
  		fprintf(stderr,"	  -GR (raw gravity) -GF (free-air gravity anomaly)\n");
  		fprintf(stderr,"	  -MT (total magnetics) -MA (magnetic anomaly) -MF (filtered magnetics)\n");
  		fprintf(stderr,"	  -TT (two-way travel time) -TC (corrected depth)\n");
  		fprintf(stderr,"	   Legs having at least the specified data are returned. [Default is any data]\n");
  		fprintf(stderr, "	-L gives a long listing (legname and datatypes). [Default is legnames only]\n");
  		fprintf(stderr, "	-V (verbose) prints number of legs found\n");
  		exit(-1);
  	}

	/* order: 1) -P argument
	          2) GMTPLUS_LIBDIR
              3) default LIBDIR
	*/
	if (path[0]) {
		sprintf(dfile, "%s/gmt_legs.d\0", path);
		sprintf(bfile, "%s/gmt_index.b\0", path);
	} else {
		envdir = getenv("GMTPLUS_LIBDIR");      
		if (envdir == NULL) {
  			sprintf (dfile, "%s/gmt_legs.d\0", LIBDIR);
  			sprintf (bfile, "%s/gmt_index.b\0", LIBDIR);
		} else {
  			sprintf (dfile, "%s/gmt_legs.d\0", envdir);
  			sprintf (bfile, "%s/gmt_index.b\0", envdir);
		}
	}

  	if ((fleg = fopen (dfile, "r")) == NULL) {
		fprintf (stderr,"gmtpluslegs: Could not open %s\n", dfile);
		exit (-1);
	}
	if ((fbin = fopen (bfile, "r")) == NULL) {
		fprintf(stderr,"gmtpluslegs: Could not open %s\n", bfile);
		exit(-1);
	}
	
	/* Make sure that WESN is in whole degrees, truncate if necessary */
	
	west = floor (west);
	east = ceil (east);
	south = floor (south);
	north = ceil (north);
	
	/* Read info about each leg */
	
	ninfo = 0;
	/* 03/14/94 jcm Note: needs to be changed; no longed bounded by 64000 */
	while (fgets (line, 512, fleg)) {
		num = sscanf (line, "%s %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
			lname, &no, &flagin[0], &flagin[1], &flagin[2], &flagin[3], 
			&flagin[4], &flagin[5], &flagin[6], &flagin[7], &flagin[8], 
			&flagin[9], &flagin[10], &flagin[11], &flagin[12], &flagin[13]);
		ptr[no] = leginfo[ninfo++] = make_info (lname, no, num-2, flagin);
	}
	fclose (fleg);

	/* Start reading the index-file */

	/* mark all legs which contain desired data */	
	while (fread ((char *)&bin, 4, 1, fbin) != 0) {
		fread ((char *)&nlegs, 4, 1, fbin);
		for (i = 0; i < nlegs; i++) {
			fread ((char *)&leg, 4, 1, fbin);
			fread ((char *)&num, 4, 1, fbin);
			for (j=0;j<num;j++) {
				fread((char *)(&flagin[j]), 4, 1, fbin);
			}
			if (inside(west, east, south, north, bin)) {
				gmt = gmtflag;
				success = TRUE;
				while ((gmt != NULL) && success) {
					for (j=0; j<num; j++) {
						if (flagin[j] == gmt->data) {
							success = TRUE;
							break;
						}
						success = FALSE; /* not in the list for this leg */
					}
					gmt = gmt->next;
				}
				if ((gmt == NULL) && success) ptr[leg]->mark = TRUE;
			}
					
		}
	}
	fclose(fbin);

	/* print out marked legs */	
	for (i = 0; i < ninfo; i++) {
		if (leginfo[i]->mark) {
			if (full) {
				printf("%s\t",leginfo[i]->legname);
				gmt = leginfo[i]->flags;
				while (gmt != NULL) {
					printf(" %s",gmt_column_name[gmt->data]);
					gmt = gmt->next;
				}
				printf("\n");
			} else {
				 (printf("%s\n", leginfo[i]->legname));
			}
			n_found++;
		}
	}
        if (gmtdefs.verbose) fprintf (stderr, "gmtpluslegs: found %d legs\n", n_found);
	exit (EXIT_SUCCESS);
}


