/*
 * mgd77_bad_detail reads a mgd77 format file such as out of the NGDC GEODAS CD-ROM
 * and looks for trouble spots.  The usage is
 *
 * mgd77_bad_detail <list_in> [-D] [-N] [-V] > output
 *
 * Input:  <list_in> = filename for a file containing filenames to check,
 * 	one per line, including the directory path if needed.
 *
 * Output:	is a report of all problems found, except 9999999 times.
 *		Use -N option to additionally output 9999999 times.
 *
 *
 * Author:	Walter H. F. Smith
 * Date:	5 August, 1994
 *
 *	Modified by Andrew Goodwillie, 12 March 1996:
 *	- see file /GDC/andrewg/gmtplus/mgd_bad_detail_info
 *
 */

#include "mgd77_error_checking.h"

void	write_out_time(const int yr, const int mo, const int dy, const int hr, const int mn, const int ndatarecs)
{


	if (hr < 10) {
		if (mn < 10) {

			if ((mo < 10) && (dy < 10)) printf("time: 0%d/0%d/19%d/0%d:0%d\t", mo, dy, yr, hr, mn);
			else if (mo < 10) printf("time: 0%d/%d/19%d/0%d:0%d\t", mo, dy, yr, hr, mn);
			else if (dy < 10) printf("time: %d/0%d/19%d/0%d:0%d\t", mo, dy, yr, hr, mn);
			else printf("time: %d/%d/19%d/0%d:0%d\t", mo, dy, yr, hr, mn);
		}
		else {
			if ((mo < 10) && (dy < 10)) printf("time: 0%d/0%d/19%d/%d:%d\t", mo, dy, yr, hr, mn);
			else if (mo < 10) printf("time: 0%d/%d/19%d/0%d:%d\t", mo, dy, yr, hr, mn);
			else if (dy < 10) printf("time: %d/0%d/19%d/0%d:%d\t", mo, dy, yr, hr, mn);
			else printf("time: %d/%d/19%d/0%d:%d\t", mo, dy, yr, hr, mn);
		}
	}
	else {
		if (mn < 10) {
			if ((mo < 10) && (dy < 10)) printf("time: 0%d/0%d/19%d/%d:0%d\t", mo, dy, yr, hr, mn);
			else if (mo < 10) printf("time: 0%d/%d/19%d/%d:0%d\t", mo, dy, yr, hr, mn);
			else if (dy < 10) printf("time: %d/0%d/19%d/%d:0%d\t", mo, dy, yr, hr, mn);
			else printf("time: %d/%d/19%d/%d:0%d\t", mo, dy, yr, hr, mn);
		}
		else {
			if ((mo < 10) && (dy < 10)) printf("time: 0%d/0%d/19%d/%d:%d\t", mo, dy, yr, hr, mn);
			else if (mo < 10) printf("time: 0%d/%d/19%d/%d:%d\t", mo, dy, yr, hr, mn);
			else if (dy < 10) printf("time: %d/0%d/19%d/%d:%d\t", mo, dy, yr, hr, mn);
			else printf("time: %d/%d/19%d/%d:%d\t", mo, dy, yr, hr, mn);
		}
	}
	if (ndatarecs <= 5) printf("\n");
}

int main (int argc, char *argv[])
{
	char line[512], name[512], errstring[128];
	BOOLEAN	error, quick_exit, verbose, data_ignore, append_time;
	BOOLEAN	output999times, output1secpairs, bad_format_or_bad_record_read;
	int	i, nheadrecs, ndatarecs, line_no, reclength, seq_no;
	FILE *fpinput, *fpmgd77;
	time_t	timenow;
	double	distance, speed;
	struct MGD77_DATA_RECORD rec, lastrec, mergerec;

	char	tempstring[16];
	int	yr, mo, dy, hr, mn, sc, idummy, returned_value;
	int	previous_yr5, previous_mo5, previous_dy5, previous_hr5, previous_mn5, previous_sc5;
	int	previous_yr4, previous_mo4, previous_dy4, previous_hr4, previous_mn4, previous_sc4;
	int	previous_yr3, previous_mo3, previous_dy3, previous_hr3, previous_mn3, previous_sc3;
	int	previous_yr2, previous_mo2, previous_dy2, previous_hr2, previous_mn2, previous_sc2;
	int	previous_yr1, previous_mo1, previous_dy1, previous_hr1, previous_mn1, previous_sc1;
	int	previous_idummy1;
	int	num999timerecords;

	fpinput = NULL;
	error = verbose = quick_exit = data_ignore = append_time = FALSE;
	output999times = FALSE, bad_format_or_bad_record_read = FALSE;
	output1secpairs = FALSE;

	for (i = 1; !error && i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case '\0':
					quick_exit = TRUE;
					break;
				case 'D':
					data_ignore = TRUE;
					break;
				case 'N':
					output999times = TRUE;
					break;
				case 'S':
					output1secpairs = TRUE;
					break;
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
		else {
			if ( (fpinput = fopen (argv[i], "r")) == NULL) {
				fprintf(stderr,"mgd77_bad_detail:  ERROR cannot open r %s\n", argv[i]);
				error = TRUE;
			}
		}
	}
	
	if (argc == 1 || error) {
		fprintf (stderr, "usage: mgd77_bad_detail <listfile> [-D] [-N] [-V]\n\n");
		if (quick_exit) exit(-1);
		fprintf (stderr, "\t<listfile> is the filename for an input file listing mgd77 files to check.\n");
		fprintf (stderr, "\t\tIf listfile is not specified, mgd77_bad_detail will look on stdin.\n");
		fprintf (stderr, "\tOutput is written to stdout.\n");
		fprintf (stderr, "\t-D will not report data range errors.  [Default reports them.]\n");
		fprintf (stderr, "\t-N will report 9999999 time field errors.  [Default does not.]\n");
		fprintf (stderr, "\t-S will report records with times less than one second since previous record.  [Default does not.]\n");
		fprintf (stderr, "\t-V will report progress.  [Default runs silently.]\n");
		exit (-1);
	}

	if (fpinput == NULL) fpinput = stdin;

	timenow = time((time_t *)0);

	while (fgets (line, 512, fpinput)) {	/* for each data file */
		line[strlen(line)-1] = '\0';
		if ( (fpmgd77 = fopen(line, "r") ) == NULL) {
			fprintf(stderr,"mgd77_bad_detail:  ERROR cannot open r %s\n", line);
			printf("\t%s could not be opened for checking.\n", line);
			continue;
		}

		nheadrecs = ndatarecs = 0;
		num999timerecords = 0;
		line_no = 0;
		strcpy(name, line);

		printf("mgd77_bad_detail report for %s\n\n", name);
		if (verbose) fprintf(stderr,"mgd77_bad_detail working on file %s", name);

		while (fgets (line, 512, fpmgd77)) {	/* read in a line of up to 512 characters from data file */
			line_no++;
			reclength = strlen(line);

			line[reclength-1] = '\0';
			reclength--;
			if (reclength == 80) {
				nheadrecs++;
				if (ndatarecs) {
					printf("\tHeader record after a data record at line %ld.\n", line_no);
				}	
				else {
					/* Check if last two columns of header are formatted correctly:  */
					if ( (sscanf(&line[78], "%d", &seq_no)) != 1 || seq_no != nheadrecs) {
						printf("\tHeader record sequence error at line %ld.\n", line_no);
					}
					else if (nheadrecs == 1 && line[0] != '1') {
						printf("\tFirst header record not marked with a 1 at line %ld.\n",  line_no);
					}
				}
			}
			else if (reclength == 120) {

				append_time = FALSE;
				bad_format_or_bad_record_read = FALSE;
				ndatarecs++;
				errstring[0] = '\0';
				if ( (mgd77_data_format_bad(line, 0, 0) ) || (mgd77_data_read_rec(line, &rec) ) ) {
					error = TRUE;
					bad_format_or_bad_record_read = TRUE;
					/*printf("\tData record format unreadable at line %ld.\t", line_no);*/
					/*continue;*/
					append_time = TRUE;
				}

				if (rec.lat_empty == 1 || rec.lon_empty == 1) {
					printf("\tData record missing lat or lon at line %ld.\t", line_no);
					continue;
				}
				if (rec.time_empty == 0 && rec.time > timenow) {
					printf("\tData record shows future time at line %ld.\t", line_no);
					continue;
				}

				/* Extract time from character string */

				strncpy(tempstring, &line[14], 2);
				tempstring[2] = '\0';
				sscanf(tempstring, "%d", &yr);

				strncpy(tempstring, &line[16], 2);
				tempstring[2] = '\0';
				sscanf(tempstring, "%d", &mo);

				strncpy(tempstring, &line[18], 2);
				tempstring[2] = '\0';
				sscanf(tempstring, "%d", &dy);

				strncpy(tempstring, &line[20], 2);
				tempstring[2] = '\0';
				sscanf(tempstring, "%d", &hr);

				strncpy(tempstring, &line[22], 5);	/* minutes x 1000 */
				tempstring[5] = '\0';
				sscanf(tempstring, "%d", &idummy);

				mn = idummy/1000;
				sc = (int)floor(0.06 * (idummy%1000));


				if (bad_format_or_bad_record_read) {

					if (output999times && ((hr == 99) || (mn == 99) || (sc == 99))) {

						num999timerecords++;
						printf("\tData record error at line %ld: Time = 9999999\n", line_no);
						continue;
					}
					if (!output999times && ((hr == 99) || (mn == 99) || (sc == 99))) {

						continue;
					}
					else if ((hr == 24) && (mn == 00)) {

						printf("\tData record error at line %ld: Time = 24:00 hours\t", line_no);
						write_out_time(yr, mo, dy, hr, mn, ndatarecs);
						printf("\n");
						continue;
					}
					else printf("\tData record format unreadable or unexpected value at line %ld.\n", line_no);
				}



				if (ndatarecs > 1) {
					/* Could check gradients here.  */
					if (lastrec.time_empty == 0 && rec.time_empty == 0) {
						if (rec.time < lastrec.time) {
							printf("\tTime decreases at line %ld.\t", line_no);
							append_time = TRUE;
						}
						if (rec.time == lastrec.time) {
							append_time = TRUE;
							if (((idummy-previous_idummy1) < 17) && idummy > 0) {

								if (output1secpairs) {
									printf("\tWarning: time less than 1 second since previous record at line %ld.\t", line_no);
									write_out_time(yr, mo, dy, hr, mn, ndatarecs);
									printf("\n");
								}
								lastrec = rec;
								previous_idummy1 = idummy;
								continue;
							}
							else if (records_are_mergable(lastrec, rec, &mergerec) ) {
								printf("\tMergable records at line %ld.\t", line_no);
							}
							else {
								printf("\tUnmergable duplicate time at line %ld.\t", line_no);
							}
						}
						distance = get_distance(rec.lat, lastrec.lat, rec.lon, lastrec.lon);
						speed = distance / (rec.time - lastrec.time);
						if (speed > MAX_VELOCITY) {


							append_time = TRUE;
							if (lastrec.qual_n != 9 || rec.qual_n != 9)
								printf("\tFlagged speed error (%.2lf m/s) at line %ld.\t", speed, line_no);
							else 
								printf("\tUnflagged speed error (%.2lf m/s) at line %ld.\t", speed, line_no);
						}
					}
				}
				if (!data_ignore && data_has_valid_range(rec, errstring) ) {
					printf("\tData range errors at line %ld:  %s\t", line_no, errstring);
					append_time = TRUE;
				}
				if (rec.time_empty != 0) {
					printf("\tRecord with no time at line %ld\t", line_no);
				}
				lastrec = rec;

			if (append_time) write_out_time(yr, mo, dy, hr, mn, ndatarecs);

			if (append_time && (ndatarecs > 5)) {

				printf("5th previous ");
				write_out_time(previous_yr5, previous_mo5, previous_dy5, previous_hr5, previous_mn5, ndatarecs);
				printf("\n");
			}

		}
		else {
			/*  A record with neither 80 nor 120 bytes is found  */
			printf("\tRecord of neither 80 nor 120 bytes at line %ld.\n", line_no);
		}


		/* find 5th previous time record */

		if (ndatarecs >= 5) {
			previous_yr5 = previous_yr4;
			previous_mo5 = previous_mo4;
			previous_dy5 = previous_dy4;
			previous_hr5 = previous_hr4;
			previous_mn5 = previous_mn4;
			previous_sc5 = previous_sc4;
		}
		if (ndatarecs >= 4) {
			previous_yr4 = previous_yr3;
			previous_mo4 = previous_mo3;
			previous_dy4 = previous_dy3;
			previous_hr4 = previous_hr3;
			previous_mn4 = previous_mn3;
			previous_sc4 = previous_sc3;
		}
		if (ndatarecs >= 3) {
			previous_yr3 = previous_yr2;
			previous_mo3 = previous_mo2;
			previous_dy3 = previous_dy2;
			previous_hr3 = previous_hr2;
			previous_mn3 = previous_mn2;
			previous_sc3 = previous_sc2;
		}
		if (ndatarecs >= 2) {
			previous_yr2 = previous_yr1;
			previous_mo2 = previous_mo1;
			previous_dy2 = previous_dy1;
			previous_hr2 = previous_hr1;
			previous_mn2 = previous_mn1;
			previous_sc2 = previous_sc1;
		}
		previous_yr1 = yr;
		previous_mo1 = mo;
		previous_dy1 = dy;
		previous_hr1 = hr;
		previous_mn1 = mn;
		previous_sc1 = sc;
		previous_idummy1 = idummy;


		}
		fclose(fpmgd77);
		if (verbose) {

			if (output999times) fprintf(stderr," - found %d records with 9999999 times\n", num999timerecords);
			else fprintf(stderr,"\n");
		}
	}
	fclose(fpinput);
	exit(0);
}

